@@ -2,7 +2,7 @@
  *
  *  BlueZ - Bluetooth protocol stack for Linux
  *
- *  Copyright (C) 2004-2009  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2004-2010  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -24,9 +24,9 @@
 #define PANU_UUID	"00001115-0000-1000-8000-00805f9b34fb"
 #define NAP_UUID	"00001116-0000-1000-8000-00805f9b34fb"
 #define GN_UUID		"00001117-0000-1000-8000-00805f9b34fb"
+#define BNEP_SVC_UUID	"0000000f-0000-1000-8000-00805f9b34fb"
 
-int bnep_init(const char *panu_script, const char *gn_script,
-		const char *nap_script);
+int bnep_init(void);
 int bnep_cleanup(void);
 
 uint16_t bnep_service_id(const char *svc);
@@ -37,5 +37,6 @@ int bnep_kill_connection(bdaddr_t *dst);
 int bnep_kill_all_connections(void);
 
 int bnep_connadd(int sk, uint16_t role, char *dev);
-int bnep_if_up(const char *devname, uint16_t id);
+int bnep_if_up(const char *devname);
 int bnep_if_down(const char *devname);
+int bnep_add_to_bridge(const char *devname, const char *bridge);
